----------------------------------------------------------------------
-- VHDL Test Bench for multiplier
----------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

library ieee;
use IEEE.STD_LOGIC_TEXTIO.all;
use STD.TEXTIO.all;

entity mult_tb is
end mult_tb;

architecture behavioural of mult_tb is
  -- Types
  type test_vect_t is array (1 to 5) of integer;  -- Test vector type
  -- Constants
  constant CLK_T : time := 100 ns;
  constant NBITS : integer := 16;
  constant TEST_VECTOR : test_vect_t := (0, 2**NBITS-1, 1, 196, 2**NBITS-2);
  -- Report files
  file results   : text open write_mode is "results.txt";
  -- Component declaration
  component mult_sum
    port(
      Clk : in  std_logic;
      X   : in  std_logic_vector(NBITS-1 downto 0);
      Y   : in  std_logic_vector(NBITS-1 downto 0);
      P   : out std_logic_vector(2*NBITS-1 downto 0)
    );
  end component;
  -- Signals
  -- -------------------------
  -- UUT replicas
  signal clk : std_logic;
  signal x   : std_logic_vector(NBITS-1 downto 0);
  signal y   : std_logic_vector(NBITS-1 downto 0);
  signal p   : std_logic_vector(2*NBITS-1 downto 0);
  --
  -- Internals
  signal do_test  : boolean := true;
  signal is_err   : boolean := false;     -- Test failure
  signal gold_p   : integer;              -- Multiplier correct result

begin

  uut : mult_sum port map(Clk => clk, X => x, Y => y, P => p);

  gen_clk : process
  begin
    while do_test loop
      clk <= '0';
      wait for CLK_T/2;
      clk <= '1';
      wait for CLK_T/2;
    end loop;
    wait; -- Wait forever
  end process;

  stim_gen_pr : process
  begin
    for i in TEST_VECTOR'range loop
      for j in TEST_VECTOR'range loop
        x <= conv_std_logic_vector(TEST_VECTOR(i), x'length);
        y <= conv_std_logic_vector(TEST_VECTOR(j), y'length);
        wait until rising_edge(clk);
      end loop;
    end loop;
    wait;
  end process stim_gen_pr;

  tb_test : process
    variable tx_loc : line;
    variable tx_str : string(1 to 4096);
  begin
    gold_p <= TEST_VECTOR(TEST_VECTOR'low)*TEST_VECTOR(TEST_VECTOR'low); --Initial value
    is_err <= false;
    -- Wait for multiplier latency:
    for latency in 1 to 2 loop 
        wait until rising_edge(clk);
    end loop;
    -- Check all results: 
    for i in TEST_VECTOR'range loop
      for j in TEST_VECTOR'range loop
        gold_p <= TEST_VECTOR(i)*TEST_VECTOR(j);
        wait until falling_edge(clk); -- check at falling edge, change if needed for timing simulation
        if (p /= conv_std_logic_vector(gold_p, p'length)) then
          is_err <= true;
          write(tx_loc, string'("ERROR!!! X=")); write(tx_loc, x);
          write(tx_loc, string'(" Y=")); write(tx_loc, y);
          write(tx_loc, string'(" P=")); write(tx_loc, p);
          write(tx_loc, string'(" "));
          tx_str(tx_loc.all'range) := tx_loc.all;
          writeline(results, tx_loc);
          Deallocate(tx_loc);
          assert (false) report tx_str severity error;
        end if;
        wait until rising_edge(clk);
      end loop;
    end loop;
    if is_err then
      report "Errors found.";
    else
      report "Simulation successful. No problems detected.";
    end if;
    do_test <= false;
    wait; -- Wait forever
  end process;
  
end;
